"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Dictionary subclass:#ResourcePack
         instanceVariableNames:'elements dependents fileName'
         classVariableNames:'Packs'
         poolDictionaries:''
         category:'System-Support'
!

ResourcePack comment:'
COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved
'!

!ResourcePack class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libview/Attic/RsrcPack.st,v 1.7 1994-07-19 16:09:13 claus Exp $
"
!

documentation
"
    to allow easy customization of smalltalk code (i.e. internationalization)
    (replaces the obsolete Resource class).
    ResourcePacks are widget class specific, meaning that every subclass of view
    has an instance of ResourcePack (called resources) which is created when
    the first instance of the view is created.
    The resourcePack consists of a mapping from strings to values, which are
    then used in labels, buttons, menus etc.
    The resourcePack initializes itself from a file found in 'resources/<className>.rs',
    where 'className' is built by the usual abbreviation mechanism (see abbrev-files).
    Conditional mappings are possible, by including lines as:
        #if <expression>
        #endif
    in the resourcefile. Example:
    file 'foo.rs':
        #if Language == #german
        'abort' 'Abbruch'
        #endif
        #if Language == #french
        'abort' 'canceller'
        #endif
    the corresponding resource-strings are accessed using:
        (resource string:'abort')
    returning the corresponding mapped string.
"
! !

!ResourcePack class methodsFor:'initialization'!

initialize
    Packs isNil ifTrue:[
        Packs := WeakArray new:30
    ].

    "ResourcePack initialize"
!

flushResources
    "forget all cached resources - needed after a style change"

    Packs := nil.
    self initialize

    "ResourcePack flushResources"
! !

!ResourcePack class methodsFor:'instance creation'!

fromFile:aFileName directory:dirName
    "get the resource definitions from a file"

    |inStream newPack temp idx|

    "first look, if not already here"

    Packs isNil ifTrue:[
        self initialize
    ].
    Packs do:[:aPack |
        aPack notNil ifTrue:[
            aPack fileName = aFileName ifTrue:[
                ^ aPack
            ]
        ]
    ].

    newPack := self new.
    inStream := Smalltalk systemFileStreamFor:('resources/' , aFileName).
    inStream isNil ifTrue:[newPack fileName:aFileName. ^ newPack].
    newPack readFromResourceStream:inStream.
    newPack fileName:aFileName.
    inStream close.

    idx := Packs identityIndexOf:nil.
    idx notNil ifTrue:[
        Packs at:idx put:newPack
    ] ifFalse:[
        temp := WeakArray new:(Packs size).
        temp replaceFrom:1 with:Packs.
        Packs := temp
    ].

    ^ newPack
!

fromFile:aFileName
    "get the resource definitions from a file"

    ^ self fromFile:aFileName directory:'resources'

    "
     ResourcePack fromFile:'SBrowser.rs'
     ResourcePack fromFile:'FBrowser.rs'
     ResourcePack fromFile:'Smalltalk.rs'
    "
!

for:aClass
    "get the resource definitions for aClass"

    |superPack|

    aClass == Object ifTrue:[
        ^ self fromFile:((Smalltalk fileNameForClass:aClass name) , '.rs')
    ].
    superPack := self for:aClass superclass.
    ^ (self fromFile:((Smalltalk fileNameForClass:aClass name) , '.rs')) merge:superPack
! !

!ResourcePack methodsFor:'merging'!

merge:anotherPack
    anotherPack associationsDo:[:assoc |
        self at:(assoc key) put:(assoc value)
    ]
! !

!ResourcePack methodsFor:'accessing'!

dependents:aCollection
    dependents := aCollection
!

dependents
    ^ dependents
!

fileName
    ^ fileName
!

fileName:aString
    fileName := aString
!

at:aKey
    ^ self at:aKey ifAbsent:[aKey]
!

at:aKey default:default
    ^ self at:aKey ifAbsent:[default]
!

name:aKey
    ^ self at:aKey ifAbsent:[aKey]
!

name:aKey default:default
    ^ self at:aKey ifAbsent:[default]
!

array:anArray
    ^ anArray collect:[:r | self at:r default:r]
!

string:s
    ^ self at:s ifAbsent:[s]
!

string:s with:arg
    ^ self string:s withArgs:(Array with:arg)
!

string:s with:arg1 with:arg2
    ^ self string:s withArgs:(Array with:arg1 with:arg2)
!

string:s withArgs:argArray
    | template expandedString idx start stop next |

    template := self at:s ifAbsent:[s].
    expandedString := ''.
    stop := template size.
    start := 1.
    [start < stop] whileTrue:[
        idx := template indexOf:$% startingAt:start.
        idx == 0 ifTrue:[
            ^ expandedString , (template copyFrom:start to:stop)
        ].
        "found a %"
        expandedString := expandedString , (template copyFrom:start to:(idx - 1)).
        next := template at:(idx + 1).
        (next == $%) ifTrue:[
            expandedString := expandedString , '%'
        ] ifFalse:[
            expandedString := expandedString , (argArray at:(next digitValue)) printString
        ].
        start := idx + 2
    ].
    ^  expandedString
! !

!ResourcePack methodsFor:'file reading'!

readFromResourceStream:inStream
    |lineString rest value ifLevel skipping l name first|

    ifLevel := 0.
    skipping := false.
    [inStream atEnd] whileFalse:[
        lineString := inStream nextLine.
        (lineString notNil and:[lineString isEmpty not]) ifTrue:[
            first := lineString at:1.
            first == $; ifFalse:[
                first == $# ifTrue:[
                    (lineString startsWith:'#if') ifTrue:[
                        skipping ifFalse:[
                            rest := lineString copyFrom:4.
                            value := Compiler evaluate:rest.
                            (value == #Error) ifTrue:[
                                Transcript show:('error in resource:' , lineString).
                            ].
                            (value == false) ifTrue:[
                                skipping := true
                            ]
                        ].
                        ifLevel := ifLevel + 1
                    ] ifFalse:[
                        (lineString startsWith:'#endif') ifTrue:[
                            ifLevel := ifLevel - 1.
                            ifLevel == 0 ifTrue:[
                                skipping := false
                            ]
                        ].
                    ].
                ] ifFalse:[
                    skipping ifFalse:[
                        lineString := lineString withoutSeparators.
                        name := nil.
                        (lineString at:1) == $' ifTrue:[
                            l := lineString indexOf:$' startingAt:2.
                            l ~~ 0 ifTrue:[
                                name := (lineString copyFrom:2 to:l-1).
                                l := l + 1
                            ]
                        ] ifFalse:[
                            l := lineString indexOfSeparatorStartingAt:1.
                            l ~~ 0 ifTrue:[
                                name := lineString copyFrom:1 to:l-1.
                            ]
                        ].
                        name notNil ifTrue:[
                            rest := lineString copyFrom:l.
                            value := Compiler evaluate:rest.
                            (value == #Error) ifTrue:[
                                Transcript show:('error in resource:' , name).
                            ].
                            self at:name put:value.
                        ]
                    ]
                ]
            ]
        ]
    ].
! !

ResourcePack initialize!
