"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#WindowSensor
         instanceVariableNames:'eventSemaphore damage mouseAndKeyboard 
                                compressMotionEvents ignoreUserInput
                                buttonState shiftDown altDown
                                metaDown controlDown exposeEventSemaphore 
                                catchExpose gotExpose
                               '
         classVariableNames:'ControlCEnabled'
         poolDictionaries:''
         category:'Interface-Support'
!

WindowSensor comment:'
COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved
'!

!WindowSensor class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libview/WindowSensor.st,v 1.8 1994-07-13 09:02:22 claus Exp $
"
!

documentation
"
    Instances of this class keep track of events and damage areas for a group of views. 
    All incoming expose rectangles and events (from Workstation) are collected here, until someone 
    gets a chance to handle them. 
    When adding an expose rectangle, WindowSensor tries to merge the rectangle with the list of 
    existing damages to minimize redrawing.

    instance variables:
        eventSemaphore          <Semaphore>     the semaphore to be signalled when an event
                                                (or damage) arrives

        damage                  <Collection>    collection of damage events

        mouseAndKeyboard        <Collection>    collection of user events

        compressMotionEvents    <Boolean>       if true, multiple motion events are
                                                compressed to one event. If false, each
                                                event is handled individual.
                                                (should be set to false when doing free-hand drawing)

        ignoreUserInput         <Boolean>       if true, key & button events are ignored
                                                (usually set to true by WindowGroup, while a
                                                 modalbox covers a view)

        buttonState                             the current state

        exposeEventSemaphore    <Semaphore>     X-special: semaphore to be signalled when
                                                expose event arrives after a copyArea.

        catchExpose             <Boolean>       true, while waiting for an expose event 
                                                (after a copyArea)

        gotExpose                               set to true, when an expose event arrives
                                                (after a copyarea)

    class variables:
        ControlCEnabled         <Boolean>       if true (which is the default) Control-C
                                                will interrupt the process handling the
                                                view.
                                                For secure stand-alone applications,
                                                this can be set to false, in which case 
                                                Control-C does NOT interrupt the process.
"
! !

!WindowSensor class methodsFor:'initialization'!

initialize
    ControlCEnabled := true
!

disableControlC
    "
     disable Control-C processing. If enabled,
     pressing CNTL-C in a view will interrupt it and bring
     its process into the debugger.
     Otherwise, CNTL-C is sent to the view like any other key.
    "

    ControlCEnabled := false
!

enableControlC
    "
     enable Control-C processing. If enabled,
     pressing CNTL-C in a view will interrupt it and bring
     its process into the debugger.
     Otherwise, CNTL-C is sent to the view like any other key.
    "

    ControlCEnabled := true
! !

!WindowSensor class methodsFor:'instance creation'!

new
    ^ self basicNew initialize
! !

!WindowSensor methodsFor:'damage'!

addDamage: aRectangle view:aView
    "Add aRectangle to the damage list.
     We try to merge incoming rectangles with the existing damage by discarding
     incoming rectangles which are contained in the existing damage or replacing
     any existing damage rectangle with incoming rectangles that completely contain it.
     There could be much more optimization here (for example joining rectangles)..."

    |count temp index newEvent|

    damage size == 0 ifTrue: [
        newEvent := WindowEvent damageFor:aView rectangle:aRectangle.
        damage := OrderedCollection with:newEvent.
        ^ self
    ].

    "
     first look, if this rectangle is already in the expose list;
     if so, dont add to queue
    "
    damage do: [:aDamage | 
        aDamage notNil ifTrue:[
            aDamage isDamage ifTrue:[
                aDamage view == aView ifTrue:[
                    ((aDamage rectangle) contains:aRectangle) ifTrue: [^self]
                ]
            ]
        ].
    ].

    "
     then look, if new rectangle contains any in the expose list;
     if so, remove the old damage
    "
    count := 0.
    1 to: damage size do:[:i |
        |aDamage|

        aDamage := damage at:i.
        aDamage notNil ifTrue:[
            aDamage isDamage ifTrue:[
                (aDamage view) == aView ifTrue:[
                    (aRectangle contains:(aDamage rectangle)) ifTrue: [ 
                        damage at:i put:nil.
                        count := count + 1
                    ]
                ]
            ]
        ]
    ].

    count > 10 ifTrue: [
        temp := OrderedCollection new:(damage size - count + 1).
        index := 1.
        damage do:[:aDamage |
            aDamage notNil ifTrue: [
                temp add: aDamage.
            ]
        ].
        damage := temp
    ].
    newEvent := WindowEvent damageFor:aView rectangle:aRectangle.
    damage add:newEvent.
!

nextDamage
    "retrieve the next damage rectangle or nil, if there is none"

    |d|

    [d isNil] whileTrue:[
        damage size == 0 ifTrue:[^ nil].
        d := damage removeFirst.
    ].
    ^ d
!

nextEvent
    "retrieve the next event or nil, if there is none"

    |e|

    [e isNil] whileTrue:[
        mouseAndKeyboard size == 0 ifTrue:[^ nil].
        e := mouseAndKeyboard removeFirst.
    ].
    ^ e
! !

!WindowSensor methodsFor:'special'!

suspendForExpose
    "asynchronous wait until a no expose arrives (after a scroll)"

    exposeEventSemaphore := Semaphore new.
    exposeEventSemaphore wait.
    exposeEventSemaphore := nil
!

catchExpose
    "start catching noExpose events (must be done after the bitblt,
     if some time may pass till the noExpose is been waited for."

    catchExpose := true.
    gotExpose := false.
!

waitForExposeFor:aView
    "synchronous wait until a no expose arrives (after a scroll)"

" "
    aView device synchronizeOutput.
" "
    [gotExpose] whileFalse:[
        aView device dispatchExposeEventFor:aView id
    ].
    catchExpose := false
!

flushEventsFor:aView
    "throw away all events for aView"

    1 to: damage size do:[:i |
        |aDamage|

        aDamage := damage at:i.
        aDamage notNil ifTrue:[
            aDamage view == aView ifTrue:[
                damage at:i put:nil
            ]
        ]
    ].
    1 to: mouseAndKeyboard size do:[:i |
        |anEvent|

        anEvent := mouseAndKeyboard at:i.
        anEvent notNil ifTrue:[
            anEvent view == aView ifTrue:[
                mouseAndKeyboard at:i put:nil
            ]
        ]
    ].
!

flushUserEvents
    "throw away all pending user events"

    mouseAndKeyboard := OrderedCollection new.
!

flushExposeEvents
    "throw away all pending expose events; this
     can be done after a full redraw (or in views, which are
     doing full redraws anly)"

    damage := OrderedCollection new.
! !

!WindowSensor methodsFor:'event processing'!

notifyEventArrival
    "an event arrived - if there is an eventsemaphore,
     signal it, to wake up any controller process"

    catchExpose == true ifTrue:[
        "dont wake up, if we are currently waiting for an expose"
        ^ self
    ].
    eventSemaphore notNil ifTrue:[
        eventSemaphore signal
    ]
!

hasDamage 
    "return true, if any exposure events are pending"

    ^ damage size ~~ 0
!

hasEvents 
    "return true, if any mouse/keyboard events are pending"

    ^ mouseAndKeyboard size ~~ 0
!

eventPending
    "return true, if either damage or events are pending"

    mouseAndKeyboard size ~~ 0 ifTrue:[^ true].
    ^ damage size ~~ 0
!

exposeX:left y:top width:width height:height view:aView
    "an expose event arrived - this is sent from the device (Display)"

    self addDamage:(Rectangle left:left top:top width:width height:height) view:aView.
    self notifyEventArrival
!

noExposeView:aView
    "an noexpose event arrived - this is sent from the device (Display)"

    gotExpose := true.
    catchExpose := false.

    exposeEventSemaphore notNil ifTrue:[
        exposeEventSemaphore signal
    ]
!

graphicExposeX:left y:top width:width height:height view:aView
    "do it now ..."
" 
    aView exposeX:left y:top width:width height:height.
    aView device synchronizeOutput
" 

    "... or better queue it ?"
" "
    self addDamage:(left @ top extent:width @ height) view:aView.
" "
!

buttonMotion:state x:x y:y view:aView
    "mouse was moved - this is sent from the device (Display)"

    |args|

    ignoreUserInput == true ifTrue:[
        ^ self
    ].
    args := Array with:state with:x with:y.

    compressMotionEvents ifTrue:[
        "
         merge with last motion
        "
        mouseAndKeyboard reverseDo:[:ev |
            ev notNil ifTrue:[
                ((ev type == #buttonMotion:x:y:) 
                and:[(ev view == aView)
                and:[(ev arguments at:1) == state]]) ifTrue:[
                    ev arguments:args.
                    ^ self
                ]
            ]
        ]
    ].
    mouseAndKeyboard
        addLast:(WindowEvent
                     for:aView
                     type:#buttonMotion:x:y:
                     arguments:args).
    self notifyEventArrival
!

buttonPress:button x:x y:y view:aView
    "mouse button was pressed - this is sent from the device (Display)"

    ignoreUserInput == true ifTrue:[
        ^ self
    ].
    mouseAndKeyboard 
        addLast:(WindowEvent 
                     for:aView
                     type:#buttonPress:x:y:
                     arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

buttonRelease:button x:x y:y view:aView
    "mouse button was released- this is sent from the device (Display)"

    ignoreUserInput == true ifTrue:[
        ^ self
    ].
    mouseAndKeyboard
        addLast:(WindowEvent
                     for:aView 
                     type:#buttonRelease:x:y:
                     arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

buttonMultiPress:button x:x y:y view:aView
    "mouse button was pressed - this is sent from the device (Display)"

    ignoreUserInput == true ifTrue:[
        ^ self
    ].
    mouseAndKeyboard
         addLast:(WindowEvent
                      for:aView
                      type:#buttonMultiPress:x:y:
                      arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

buttonShiftPress:button x:x y:y view:aView
    "mouse button was pressed - this is sent from the device (Display)"

    ignoreUserInput == true ifTrue:[
        ^ self
    ].
    mouseAndKeyboard
        addLast:(WindowEvent
                     for:aView
                     type:#buttonShiftPress:x:y:
                     arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

keyPress:key x:x y:y view:aView
    "key was pressed - this is sent from the device (Display)"

    |xlatedKey group process|

    xlatedKey := aView device translateKey:key.
    ((xlatedKey == #Ctrlc) and:[ControlCEnabled]) ifTrue:[
        "
         Special handling for Cntl-C: interrupt the underlying process.

         cannot halt here (this would stop the event-dispatcher),
         but instead interrupt the underlying process and have it
         perform the userInterrupt in the interrupt-method.
        "
        group := aView windowGroup.
        group notNil ifTrue:[
            process := group process.
            process notNil ifTrue:[
                process interruptWith:[process userInterrupt]
            ]
        ].
        ^ self
    ].
    ignoreUserInput == true ifTrue:[
        ^ self
    ].
    mouseAndKeyboard
        addLast:(WindowEvent
                     for:aView
                     type:#keyPress:x:y:
                     arguments:(Array with:xlatedKey with:x with:y)).
    self notifyEventArrival
!

pointerEnter:state x:x y:y view:aView
    "mouse cursor was moved into the view - this is sent from the device (Display)"

    mouseAndKeyboard
        addLast:(WindowEvent
                  for:aView
                  type:#pointerEnter:x:y:
                  arguments:(Array with:state with:x with:y)).
    self notifyEventArrival
!

pointerLeave:state view:aView
    "mouse cursor was moved out of the view - this is sent from the device (Display)"

    mouseAndKeyboard
         addLast:(WindowEvent
                     for:aView
                     type:#pointerLeave: 
                     arguments:(Array with:state)).
    self notifyEventArrival
!

configureX:x y:y width:w height:h view:aView
    damage
        addLast:(WindowEvent
                     for:aView
                     type:#configureX:y:width:height:
                     arguments:(Array with:x with:y with:w with:h)).
    self notifyEventArrival
!

focusInView:aView
    "view got input focus - this is sent from the device (Display)"

    mouseAndKeyboard
         addLast:(WindowEvent
                     for:aView
                     type:#focusIn 
                     arguments:#()).
    self notifyEventArrival
!

focusOutView:aView
    "view lost input focus - this is sent from the device (Display)"

    mouseAndKeyboard
         addLast:(WindowEvent
                     for:aView
                     type:#focusOut 
                     arguments:#()).
    self notifyEventArrival
!

mappedView:aView
    "view was mapped (from window manager)"

    damage
         addLast:(WindowEvent
                     for:aView
                     type:#mapped 
                     arguments:#()).
    self notifyEventArrival
!

unmappedView:aView
    "view was unmapped (from window manager)"

    damage
         addLast:(WindowEvent
                     for:aView
                     type:#unmapped 
                     arguments:#()).
    self notifyEventArrival
!

terminateView:aView
    "view should terminate (from window manager)"

    self flushEventsFor:aView.
    damage
         addLast:(WindowEvent
                     for:aView
                     type:#terminate 
                     arguments:#()).
    self notifyEventArrival
!

saveAndTerminateView:aView
    "view should save & terminate (from window manager)"

    self flushEventsFor:aView.
    damage
         addLast:(WindowEvent
                     for:aView
                     type:#saveAndTerminate 
                     arguments:#()).
    self notifyEventArrival
!

destroyedView:aView
    "view was destroyed (from window manager)"

    "at this time, the view is already gone; remove
     all pending events for this one ..."

    self flushEventsFor:aView.
    damage
         addLast:(WindowEvent
                     for:aView
                     type:#destroyed
                     arguments:#()).
    self notifyEventArrival
! !

!WindowSensor methodsFor:'initialization'!

initialize
    "initialize the event queues to empty"

    damage := OrderedCollection new.
    mouseAndKeyboard := OrderedCollection new.
    gotExpose := true.
    catchExpose := false.
    compressMotionEvents := true.
    ignoreUserInput := false.
! !

!WindowSensor methodsFor:'accessing'!

ignoreUserInput:aBoolean
    ignoreUserInput := aBoolean
!

ignoreUserInput
    ^ ignoreUserInput
!

eventSemaphore:aSemaphore
    "set the semaphore used to signal event arrival"

    eventSemaphore := aSemaphore
!

eventSemaphore
    "return the semaphore used to signal event arrival"

    ^ eventSemaphore
!

compressMotionEvents:aBoolean
    compressMotionEvents := aBoolean
! !
